package org.netkernel.module.standard.builtin.branchmerge;

import java.util.*;
import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.module.standard.*;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.w3c.dom.*;

public class BranchMergeOverlay extends ConfiguredOverlayImpl
{
	private static final PairList VALIDATION = new PairList(5);
	static
	{	
		VALIDATION.put("/*[name()!='config']","MSG_STD_BM_1");
		VALIDATION.put("/config/*[name()!='overlay' and name()!='branch']", "MSG_STD_BM_2");
		VALIDATION.put("/config/overlay/attribute::*[name()!='name' and name()!='next']", "MSG_STD_BM_3");
		VALIDATION.put("/config/branch/attribute::*[name()!='name' and name()!='overlay' and name()!='targets']", "MSG_STD_BM_4");
	}
	
	public BranchMergeOverlay()
	{	declareThreadSafe();
		declareUnhandledExceptionsExpired(false);
	}
	
	protected IConfig createConfig(INKFRequestContext aContext) throws Exception
	{
		Document configDoc = Utils.sourceAsDOMAvoidingTransrept(aContext, "param:config");
		
		ISpaceMeta delegateMeta=getDelegateMeta(aContext);
		BranchMergeConfig config = new BranchMergeConfig(configDoc, getDelegateSpace(), delegateMeta, aContext);
		return config;
	}
	
	protected void destroyConfig(IConfig aConfig) throws Exception
	{
		//TODO clean up spaces
	}	
	
	public String toString()
	{	return "BranchMergeOverlay";
	}

	public String getName()
	{	return toString();
	}
	
	public IKernel getKernel()
	{	return super.getKernel();
	}

	
	
	
	private class BranchMergeConfig implements IConfig
	{
		private ISpace mDelegateSpace;
		private ISpaceElements mElements;
		private Map<String, IMetaRepresentation> mElementMetasById;
		private Map<String, BranchMergeSpace> mSpacesByName;
		private Map<String,Branch> mBranchesById;
		
		public BranchMergeConfig(Document aConfig, ISpace aDelegateSpace, ISpaceMeta aDelegateMeta, INKFRequestContext aContext) throws Exception
		{
			ILogger l=aContext.getKernelContext().getKernel().getLogger();
			
			mDelegateSpace=aDelegateSpace;
			
			//construct space elements, and get all elements meta
			ISpaceElements delegateElements=aDelegateMeta.getElements();			
			mElementMetasById = new LinkedHashMap<String, IMetaRepresentation>();
			List<String> unreferencedTargets = new ArrayList<String>();
			for (int i=0; i<delegateElements.size(); i++)
			{	IIdentifier elementId=delegateElements.getIdentifier(i);
				if (elementId!=null)
				{	String elementIdString = elementId.toString();
					unreferencedTargets.add(elementIdString);
					IMetaRepresentation meta=aContext.meta(elementIdString);
					mElementMetasById.put(elementIdString, meta);
				}
			}
			// build our list of elements
			TupleList tl=new TupleList(3,delegateElements.size());
			for (int i=0; i<delegateElements.size(); i++)
			{	tl.put(delegateElements.getIdentifier(i),aDelegateSpace,BranchMergeOverlay.this);
			}
			mElements=new SpaceElementsImpl(tl);
			
			//validate
			PairList validation=FastSchematron.validate(aConfig.getDocumentElement(), VALIDATION);
			if (validation.size()>0)
			{	String m = FastSchematron.formatValidationError(validation,aContext.getKernelContext().getKernel().getLogger());			
				throw Utils.createFormattedException("EX_STD_BM_CONFIG", "MSG_RAW", aContext.getKernelContext().getKernel().getLogger(), m);
			}
			
			XMLReadable r = new XMLReadable(aConfig.getDocumentElement());
			List<Element> overlays=(List<Element>)(List)r.getNodes("overlay");
			Map<String, Element> namedOverlays = new HashMap<String, Element>();
			for (Element overlay: overlays)
			{	String overlayName=overlay.getAttribute("name");
				namedOverlays.put(overlayName,overlay);
			}
			
			//construct spaces and test for circularity, all paths lead to delegate space
			mSpacesByName = new HashMap<String, BranchMergeSpace>();
			while(mSpacesByName.size()<namedOverlays.size())
			{	boolean progress=false;
				for (String overlayName: namedOverlays.keySet())
				{
					if (mSpacesByName.get(overlayName)==null)
					{
						Element overlayConfig=namedOverlays.get(overlayName);
						String next=overlayConfig.getAttribute("next");
						ISpace nextSpace=null;
						if (next.equals(""))
						{	nextSpace=aDelegateSpace;
						}
						else
						{	nextSpace=mSpacesByName.get(next);
						}
						if (nextSpace!=null)
						{	BranchMergeSpace constructedSpace = constructSpace(overlayConfig, nextSpace, aDelegateSpace, aContext, overlayName);
							mSpacesByName.put(overlayName, constructedSpace);
							progress=true;
						}
					}
				}
				if (!progress)
				{	throw Utils.createFormattedException("EX_STD_BM_CONFIG", "MSG_STD_BM_NO_ROUTE", l);
				}
			}
			
			//construct branches
			Set<String> usedOverlays = new HashSet<String>();
			List<Element> branches=(List<Element>)(List)r.getNodes("branch");
			mBranchesById = new HashMap<String,Branch>();
			for (Element branch: branches)
			{	String branchName=branch.getAttribute("name");
				String overlayName=branch.getAttribute("overlay");
				ISpace space;
				if (overlayName.length()==0)
				{	space=aDelegateSpace;
				}
				else
				{	BranchMergeSpace bms=mSpacesByName.get(overlayName);
					if (bms==null)
					{	throw Utils.createFormattedException("EX_STD_BM_CONFIG", "MSG_STD_BM_NO_OVERLAY", l,overlayName,branchName);
					}
					space=bms;
					usedOverlays.add(overlayName);
					ISpace s=bms.getNextSpace();
					while (s instanceof BranchMergeSpace)
					{	bms=(BranchMergeSpace)s;
						usedOverlays.add(bms.getName());
						s=bms.getNextSpace();
					}
				}
				
				String targetsString=branch.getAttribute("targets");
				String[] targets=Utils.splitString(targetsString, ", ");
				if (targets.length!=0)
				{	int warnings=0;
					for (int i=0; i<targets.length; i++)
					{	IMetaRepresentation elementMeta=mElementMetasById.get(targets[i]);
						if (elementMeta==null)
						{	ILogger logger=getKernel().getLogger();
							logger.log(ILogger.WARNING, this, "MSG_STD_BM_NO_TARGET", targets[i],branchName);
							targets[i]=null;
							warnings++;
						}
					}
					if (warnings>0)
					{	//filter out nulls from array
						String[] targets2 = new String[targets.length-warnings];
						int j=0;
						for (int i=0; i<targets.length; i++)
						{	String target=targets[i];
							if (target!=null)
							{	targets2[j++]=target;
							}
						}
						targets=targets;
					}
				}
				else
				{	//use all remaining targets
					targets=unreferencedTargets.toArray(targets);
					unreferencedTargets.clear();
				}
				
				Branch b = new Branch(branchName,space,targets);
				for (int i=0; i<targets.length; i++)
				{	String target=targets[i];
					mBranchesById.put(target,b);
					unreferencedTargets.remove(target);
				}
			}
			
			//test all overlays are used by branches
			for (String overlayName: mSpacesByName.keySet())
			{	if (!usedOverlays.contains(overlayName))
				{	l.log(ILogger.WARNING, this, "MSG_STD_BM_UNUSED_OVERLAY",overlayName);
				}
			}

		}
		
		private BranchMergeSpace constructSpace(Element aConfig, ISpace aNextSpace, ISpace aDelegateSpace, INKFRequestContext aContext, String aName) throws Exception
		{
			//construct overlay
			IKernel k=aContext.getKernelContext().getKernel();
			ClassLoader cl= aContext.getKernelContext().getOwningSpace().getClassLoader();
			BranchMergeOverlay owner = BranchMergeOverlay.this;
			IStandardEndpoint overlay=SpaceFactory.buildEndpoint(aConfig, XMLUtils.getPathFor(aConfig), cl, k);
			overlay.setParameterRaw(IStandardOverlay.PARAM_SPACE, aNextSpace,k);
			BranchMergeSpace space= new BranchMergeSpace(overlay,aDelegateSpace,aNextSpace,aName,owner);
			space.onCommissionSpace(k);
			space.postCommissionSpace();
			return space;
		}
		

		public ISpaceElements getElements(INKFRequestContext context) throws Exception
		{	return mElements;
		}

		public IMetaRepresentation getElementMeta(String elementId)
		{	return mElementMetasById.get(elementId);
		}

		public String getResolvedElement(INKFRequestReadOnly aToResolve) throws Exception
		{	String result=null;
			for (Map.Entry<String, IMetaRepresentation> entry: mElementMetasById.entrySet())
			{	IMetaRepresentation meta=entry.getValue();
				if (meta instanceof IEndpointMeta)
				{	if (MetadataUtils.match((IEndpointMeta)meta, aToResolve))
					{	result = entry.getKey();
						break;
					}
				}
			}
			return result;
		}

		public void onRequest(String aElementId, INKFRequestContext aContext) throws Exception
		{
			Branch b=mBranchesById.get(aElementId);
			ISpace delegateSpace;
			if (b==null)
			{	delegateSpace=mDelegateSpace;
			}
			else
			{	delegateSpace=b.getSpace();
			}
			INKFRequest subReq=aContext.getThisRequest().getIssuableClone();
			subReq.injectRequestScope(delegateSpace);
			INKFResponseReadOnly<?> resp=aContext.issueRequestForResponse(subReq);
			aContext.createResponseFrom(resp);
		}

		public ISpace[] getSpaces(INKFRequestContext context) throws Exception
		{
			return IStandardEndpoint.NO_SPACES;
		}
		
	}

	//TODO this Branch class isn't needed as only space is used
	private static class Branch
	{
		private final String mName;
		private final ISpace mSpace;
		private final String[] mTargets;
		
		public Branch(String aName, ISpace aSpace, String[] aTargets)
		{	mName=aName;
			mSpace=aSpace;
			mTargets=aTargets;
		}
		
		public ISpace getSpace()
		{	return mSpace;
		}
	}

	
}
